import json
from asyncio import ensure_future, sleep
from logging import Handler, Logger
from typing import Literal, Union

from pydantic import BaseModel, BaseSettings, Field

DEFAULT_DURATION_SECONDS = 60


class LogLevelListenerConfig(BaseSettings):
    """Pydantic settings object to configure a log level listener."""

    FALLBACK_LOG_LEVEL: str = Field(
        description="Log level to reset to after the duration specified in the LogLevelRequestMessage",
        default="INFO",
    )


class LogLevelRequestMessage(BaseModel):
    """Message representing a request to temporarily change the log level."""

    level: Literal["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"] = Field(
        description="a valid Python logging level (CRITICAL, ERROR, WARNING, INFO, DEBUG)"
    )
    duration: float = Field(
        description="length of time in seconds before the log level is reset; "
        "setting this value to 0 or a negative integer will cause the log level to not be reset",
        default=DEFAULT_DURATION_SECONDS,
        ge=0,
    )


class BaseAsyncLogLevelListener:
    """
    Log level listener that changes the log level of the logging handler with name="stdout".

    Changes the log level of the stdout logging handler (or the logger itself if no stdout logging
    handler is found) for a specified duration. These parameters are provided via a LogLevelRequestMessage.
    """

    logger: Logger
    config: LogLevelListenerConfig

    def __init__(self, logger: Logger, config: LogLevelListenerConfig):
        self.logger = logger
        self.config = config

    def _get_stdout_handler_or_logger(self) -> Union[Handler, Logger]:
        for handler in self.logger.handlers:
            if handler.get_name() == "stdout":
                return handler
        return self.logger

    async def reset_log_level(
        self,
        handler_or_logger: Union[Handler, Logger],
        level: Union[str, int],
        delay: float,
    ) -> None:
        """Reset the log level after a certain delay."""
        await sleep(delay)
        self.logger.info(f"Resetting log level to {level}")
        handler_or_logger.setLevel(level)

    def set_log_level(
        self, level: str, duration: float = DEFAULT_DURATION_SECONDS
    ) -> None:
        """Set the log level and starts and async task to reset it later."""
        self.logger.info(f"Setting log level to {level}")
        handler_or_logger = self._get_stdout_handler_or_logger()
        handler_or_logger.setLevel(level)

        if duration > 0:
            self.logger.info(f"Log level will be reset in {duration} seconds")
            ensure_future(
                self.reset_log_level(
                    handler_or_logger, self.config.FALLBACK_LOG_LEVEL, duration
                )
            )
        else:
            self.logger.warning("Log level will not be automatically reset")

    async def handle_log_level_change_message(
        self, json_message: Union[str, bytes]
    ) -> LogLevelRequestMessage:
        body = json.loads(json_message)
        request = LogLevelRequestMessage(**body)
        self.set_log_level(level=request.level, duration=request.duration)
        return request



"""This module defines a Logger with a .trace method for logging at the TRACE level.

Import this module to enable trace logging.
"""
import logging
from typing import Mapping, Union

TRACE = 5

logging.addLevelName(TRACE, "TRACE")


class TraceLogger(logging.Logger):
    """A logging.Logger subclass that can log at the trace severity level."""

    def trace(
        self,
        msg: object,
        *args: object,
        exc_info: bool = False,
        stack_info: bool = False,
        stacklevel: int = 0,
        extra: Union[Mapping[str, object], None] = None,
    ) -> None:
        """Log at the trace severity level.

        Usage is the same as other logging methods, such as Logger.info.
        """
        self.log(
            TRACE,
            msg,
            *args,
            exc_info=exc_info,
            stack_info=stack_info,
            stacklevel=stacklevel,
            extra=extra,
        )



import json
from json import JSONDecodeError
from logging import Logger
from typing import Optional

from nats.aio.client import Client as NATSClient
from nats.aio.msg import Msg
from pydantic import Field, ValidationError

from powerflex_logging_utilities.log_level_listener import (
    BaseAsyncLogLevelListener,
    LogLevelListenerConfig,
)
from powerflex_logging_utilities.log_level_listener.format_exception import (
    format_exception,
)


class NatsLogLevelListenerConfig(LogLevelListenerConfig):
    """Pydantic settings object to configure a NATS log level listener."""

    NATS_LOG_LEVEL_LISTENER_SUBJECT: str = Field(
        description="NATS subject that will receive LogLevelRequestMessage messages"
    )


class AsyncNatsLogLevelListener(BaseAsyncLogLevelListener):
    """
    NATS listener that changes the log level of the logging handler with name="stdout".

    Changes the log level of the stdout logging handler (or the logger itself if no stdout logging
    handler is found) for a specified duration. These parameters are provided via a LogLevelRequestMessage.
    """

    nats_client: NATSClient
    logger: Logger
    config: NatsLogLevelListenerConfig

    def __init__(
        self,
        nats_client: NATSClient,
        logger: Logger,
        config: NatsLogLevelListenerConfig,
    ):
        self.nats_client = nats_client
        super().__init__(logger, config)

    async def async_init(self) -> None:
        self.logger.info(
            "Log Level Listener subscribing to subject %s",
            self.config.NATS_LOG_LEVEL_LISTENER_SUBJECT,
        )
        await self.nats_client.subscribe(
            subject=self.config.NATS_LOG_LEVEL_LISTENER_SUBJECT,
            cb=self.handle_log_level_change_nats_message,
        )

    @staticmethod
    async def create(
        nats_client: NATSClient,
        logger: Logger,
        config: Optional[NatsLogLevelListenerConfig] = None,
    ) -> "AsyncNatsLogLevelListener":
        if config is None:
            config = NatsLogLevelListenerConfig(
                # type: ignore
            )
        log_level_listener = AsyncNatsLogLevelListener(
            nats_client=nats_client, logger=logger, config=config
        )
        logger.info(
            f"Log Level Listener initialized with config {config}", extra=config.dict()
        )
        await log_level_listener.async_init()
        return log_level_listener

    async def handle_log_level_change_nats_message(self, msg: Msg) -> None:
        try:
            request = await self.handle_log_level_change_message(msg.data)
            if msg.reply:
                response = {"result": f"Log level set to {request.level}"}
                await self.nats_client.publish(msg.reply, json.dumps(response).encode())
        except ValidationError as ex:
            if msg.reply:
                response = {"error": format_exception(ex)}
                await self.nats_client.publish(msg.reply, json.dumps(response).encode())
        except JSONDecodeError as ex:
            if msg.reply:
                response = {"error": "Error when parsing JSON\n" + format_exception(ex)}
                await self.nats_client.publish(msg.reply, json.dumps(response).encode())



import sys
import traceback
from typing import Optional


def format_exception(exception: Optional[BaseException] = None) -> str:
    if exception is None:
        _, exception, _ = sys.exc_info()
    if exception is None:
        return "No exception"
    return "".join(traceback.format_exception_only(type(exception), exception))



import logging
import logging.handlers
import os
import sys
from typing import Any, Collection, Dict, Optional, TextIO, Type, Union, cast

from powerflex_logging_utilities.default_log_format import DEFAULT_LOG_FORMAT
from powerflex_logging_utilities.json_formatter import JsonFormatter

DEFAULT_LOGFILE_MAX_BYTES = 1000 * 1000 * 10  # 10 megabytes
DEFAULT_LOGFILE_BACKUP_COUNT = 25


def min_log_level(level1: Union[str, int], level2: Optional[Union[str, int]]) -> int:
    if isinstance(level1, str):
        level1 = cast(int, logging.getLevelName(level1))
    if not isinstance(level2, int):
        level2 = cast(int, logging.getLevelName(level2 or "CRITICAL"))
    return min([level1, level2])


def add_stream_handler(
    logger_instance: logging.Logger,
    log_level: Union[str, int],
    formatter: Type[logging.Formatter],
    formatter_kwargs: Optional[Dict[str, Any]] = None,
    log_format: str = DEFAULT_LOG_FORMAT,
    stream: TextIO = sys.stdout,
) -> None:
    if formatter_kwargs is None:
        formatter_kwargs = {}
    log_handler = logging.StreamHandler(stream=stream)
    log_handler.set_name("stdout")
    log_handler.setFormatter(formatter(fmt=log_format, **formatter_kwargs))
    log_handler.setLevel(log_level)
    logger_instance.addHandler(log_handler)


def add_file_handler(
    logger_instance: logging.Logger,
    log_level: Union[str, int],
    filename: str,
    max_bytes: int,
    backup_count: int,
    formatter: Type[logging.Formatter],
    formatter_kwargs: Optional[Dict[str, Any]] = None,
    log_format: str = DEFAULT_LOG_FORMAT,
) -> None:
    """Add a file handler to a Logger so it logs to a file.

    Creates the log file directory if it doesn't exist.
    """
    if formatter_kwargs is None:
        formatter_kwargs = {}

    os.makedirs(os.path.dirname(filename), exist_ok=True)

    log_handler = logging.handlers.RotatingFileHandler(
        filename,
        maxBytes=max_bytes,
        backupCount=backup_count,
    )
    log_handler.setFormatter(formatter(fmt=log_format, **formatter_kwargs))
    log_handler.setLevel(log_level)
    logger_instance.addHandler(log_handler)


def init_logger(
    log_level: Union[str, int],
    file_log_level: Optional[Union[str, int]],
    filename: Optional[str],
    logger_instance: Union[logging.Logger, str] = "",
    max_bytes: int = DEFAULT_LOGFILE_MAX_BYTES,
    backup_count: int = DEFAULT_LOGFILE_BACKUP_COUNT,
    stream: TextIO = sys.stdout,
    formatter: Type[logging.Formatter] = JsonFormatter,
    formatter_kwargs: Optional[Dict[str, Any]] = None,
    log_format: str = DEFAULT_LOG_FORMAT,
) -> None:
    """Configure a logger to log to both the given stream and filename with the given formatter.

    logger_instance - Configure this Logger object.
        If a string, configure logging.getLogger(logger_instance)
    """
    if isinstance(logger_instance, str):
        logger_instance = logging.getLogger(logger_instance)

    # We need the minimum log level of the two so that we don't accidentally
    # set a logger to log at a higher level than is expected by either the
    # stdout or file handlers.
    min_level = min_log_level(log_level, file_log_level)

    logger_instance.setLevel(min_level)

    add_stream_handler(
        logger_instance,
        log_level,
        formatter,
        formatter_kwargs,
        log_format,
        stream=stream,
    )
    if not (file_log_level is None or filename is None):
        add_file_handler(
            logger_instance,
            file_log_level,
            filename,
            max_bytes,
            backup_count,
            formatter,
            formatter_kwargs,
            log_format,
        )


def init_loggers(
    loggers: Collection[Union[logging.Logger, str]],
    log_level: Union[str, int],
    file_log_level: Optional[Union[str, int]],
    filename: Optional[str],
    max_bytes: int = DEFAULT_LOGFILE_MAX_BYTES,
    backup_count: int = DEFAULT_LOGFILE_BACKUP_COUNT,
    stream: TextIO = sys.stdout,
    formatter: Type[logging.Formatter] = JsonFormatter,
    formatter_kwargs: Optional[Dict[str, Any]] = None,
    log_format: str = DEFAULT_LOG_FORMAT,
    info_logger: Optional[Union[logging.Logger, str]] = None,
) -> None:
    """Configure loggers to log to both the given stream and filename with the given formatter.

    loggers - Configure these Logger objects.
        If a string is passed instead of a Logger, configure logging.getLogger(logger_instance)

    info_logger - If not None, use it to log the logging configuration after setting up loggers.
        If a string is passed instead of a Logger, use logging.getLogger(info_logger)
    """
    for logger_instance in loggers:
        init_logger(
            log_level,
            file_log_level,
            filename,
            logger_instance,
            max_bytes,
            backup_count,
            stream,
            formatter,
            formatter_kwargs,
            log_format,
        )

    if info_logger is None:
        return

    if isinstance(info_logger, str):
        info_logger = logging.getLogger(info_logger)

    info_logger.info("Logging at level %s", log_level)
    if file_log_level is None or filename is None:
        info_logger.info("Not logging to a file")
    else:
        info_logger.info(
            "Logging at level %s to file %s",
            file_log_level,
            filename,
            extra={
                "config": dict(
                    maxBytes=max_bytes,
                    backupCount=backup_count,
                )
            },
        )



import logging
from typing import Any, Callable


def forbid_logging_with_logging_toplevel() -> None:
    """Forbid the use of logging using the top-level logging functions in the \
    logging module, such as logging.info and logging.critical.

    This should help you only use the non-root logger.

    When using a JSON logger, using logging functions from the logging module
    top-level could cause non-JSON formatted logs to appear alongside JSON
    logs. This is because the logging module configures itself when used like
    this for the first time.
    """

    def make_forbidden_log_function(name: str) -> Callable[[Any], None]:
        def forbidden_log_function(*_args: Any, **_kwargs: Any) -> None:
            raise RuntimeError(
                f"Do not use logging.{name}, use a named logger such as traffic_manager.log.logger instead"
            )

        return forbidden_log_function

    log_functions = [
        "info",
        "debug",
        "error",
        "warning",
        "warn",
        "critical",
        "log",
    ]
    for log_function in log_functions:
        setattr(logging, log_function, make_forbidden_log_function(log_function))



from typing import Any, Dict, Optional

# Wait for an update or write type stubs
from pythonjsonlogger import jsonlogger  # type: ignore

from powerflex_logging_utilities.default_log_format import DEFAULT_LOG_FORMAT


class JsonFormatter(jsonlogger.JsonFormatter):  # type: ignore
    """A custom JSON formatter.

    - Replaces the "levelname" field with "severity".
    - Replaces non-string keys in the log record with their __str__ representation
    """

    def __init__(self, fmt: Optional[str] = DEFAULT_LOG_FORMAT, **kwargs: Any):
        super().__init__(fmt=fmt, **kwargs)

    def process_log_record(self, log_record: Dict[str, Any]) -> Any:
        log_record["severity"] = log_record["levelname"]
        del log_record["levelname"]

        for key in log_record.copy():
            if not isinstance(key, str):
                log_record[str(key)] = log_record[key]
                del log_record[key]

        return super().process_log_record(log_record)



import os.path

from .default_log_format import DEFAULT_LOG_FORMAT
from .json_formatter import JsonFormatter
from .trace_logger import TRACE, TraceLogger

with open(
    os.path.normpath(os.path.join(__file__, "../", "VERSION")), encoding="utf-8"
) as versionfile:
    __version__ = versionfile.read().strip()



from os.path import dirname, join

from setuptools import find_packages, setup


def read(name, **kwargs):
    with open(
        join(dirname(__file__), name), encoding=kwargs.get("encoding", "utf8")
    ) as openfile:
        return openfile.read()


setup(
    name="powerflex-logging-utilities",
    packages=find_packages("src"),
    package_dir={"": "src"},
    zip_safe=True,
    version=read("src/powerflex_logging_utilities/VERSION").strip(),
    package_data={"powerflex_logging_utilities": ["VERSION", "py.typed"]},
    include_package_data=True,
    license="MIT License",
    description="Helpful code for logging in Python",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    url="https://github.com/edf-re/powerflex_logging_utilities_py",
    project_urls={
        "Issue Tracker": "https://github.com/edf-re/powerflex_logging_utilities_py/issues",
    },
    keywords=["NATS", "NATS request", "aiodebug", "async", "JSON logging"],
    # Only 3.8+
    python_requires=">2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*, !=3.6.*, !=3.7.*",
    install_requires=[
        "aiodebug>=2.3.0",
        "python-json-logger>=2.0.4",
    ],
    extras_require={
        "pydantic": ["pydantic>=1.9"],
        "nats-and-pydantic": ["nats-py>=2", "pydantic"],
    },
    classifiers=[
        "Intended Audience :: Developers",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Utilities",
    ],
)



from logging import INFO, WARN, Logger

# For some reason, mypy doesn't detect this package's type hints
import aiodebug.log_slow_callbacks  # type: ignore


def log_slow_callbacks(
    logger: Logger,
    slow_async_task_threshold_sec: float = 0.15,
    very_slow_async_task_threshold_sec: float = 0.5,
) -> None:
    """Log at INFO or WARN severity levels when an async task runs too slowly.

    slow_async_task_threshold_sec - INFO log if an async task runs longer than
        this many seconds. Python's default is 0.1sec, but our default is
        higher.

    very_slow_async_task_threshold_sec - WARN log if an async task runs longer
        than this many seconds.
    """

    def on_slow_callback(task_name: str, duration: float) -> None:
        level = INFO
        if duration > very_slow_async_task_threshold_sec:
            level = WARN
        logger.log(
            level,
            "Executing task %s blocked async loop for %s seconds",
            task_name,
            duration,
            extra={"task_name": task_name, "duration": duration},
        )

    aiodebug.log_slow_callbacks.enable(
        slow_async_task_threshold_sec, on_slow_callback=on_slow_callback
    )
